#!/usr/bin/python

# Solar Monitor for SRNE ML2440 etc by Lindsay Crawford - 2018/09/01
# Free to use, I make no guarantee that it will work with your controller 
# and by using this software you agree to not hold me accountable in the 
# unlikely event that this causes damage to you or your solar controller, 
# Raspberry PI or any other attached devices.

import time, datetime, serial, os
import sqlite3
import RPi.GPIO as GPIO
import os
# Import Pymodbus
from pymodbus.client.sync import ModbusSerialClient as ModbusClient 
from pymodbus.transaction import ModbusRtuFramer
from shutil import copyfile
# Import the ADS1x15 module.
import Adafruit_ADS1x15

# Set initial GPIO states
# 23 used for relay to reboot internet if it goes down
# 13 Red, 19 Orange, 26 Green LED (390ohm resistors for all) to indicate charge status for the day
# 25 used for reed relay to turn on the shunt inputs to the ADS1x15 module. This prevents damage that seems 
# to happen to the ADS1x15 if the Raspberry/ADS1x15 is powered off and the inputs still have voltage
GPIO.setwarnings(False)
GPIO.setmode(GPIO.BCM)
GPIO.setup(23, GPIO.OUT)
GPIO.setup(5, GPIO.OUT)
GPIO.setup(6, GPIO.OUT)
GPIO.setup(13, GPIO.OUT)
GPIO.setup(19, GPIO.OUT)
GPIO.setup(26, GPIO.OUT)
GPIO.setup(25, GPIO.OUT)
GPIO.output(23, True)
GPIO.output(5, False)
GPIO.output(6, False)
GPIO.output(13, False)
GPIO.output(19, False)
GPIO.output(26, False)
GPIO.output(25, False)

# LED values at which they should switch on
yellowLED = -10
whiteLED = -2
greenLED = 2
blueLED = 10

# Initial values for internet connection checking
hostname = "google.com"
noInternetCount = 0
lastInternetReset = datetime.datetime(1900, 1, 1, 1, 1, 1)
# After 15 loops with no internet reset relay on pin 23 to reboot router
maxNoInternet = 15

# Sleep to make sure all drivers are loaded, not sure if actually necessary
# Flash the LEDs for a bit
for i in range(2):
	GPIO.output(6, False)
	GPIO.output(5, True)
	time.sleep(1)
	GPIO.output(5, False)
	GPIO.output(6, True)
	time.sleep(1)
	GPIO.output(6, False)
	GPIO.output(13, True)
	time.sleep(1)
	GPIO.output(13, False)
	GPIO.output(19, True)
	time.sleep(1)
	GPIO.output(19, False)  
	GPIO.output(26, True)
    time.sleep(1)
	GPIO.output(26, False) 
	GPIO.output(19, True)
    time.sleep(1)
	GPIO.output(19, False)
    GPIO.output(13, True) 
	time.sleep(1)
	GPIO.output(13, False) 
	GPIO.output(6, True) 
	time.sleep(1)
GPIO.output(6, False)
GPIO.output(5, True)
time.sleep(1)
GPIO.output(5, False)

# Create an ADS1115 ADC (16-bit) instance and set initial values for using and storing data from this
# External board used to measure current through a shunt
adc = Adafruit_ADS1x15.ADS1115()
GAIN=16
# Set reed relay on to allow reading current through shunt
GPIO.output(25, True)
time.sleep(2)
directTotalAh = 0.0
directTotalWh = 0.0
directMaxAmps = 0.0
prevDischargeAh = 0.0
prevChargeAh = 0.0
start = time.time()
directAmps = 0.0
daytime = False

# Create ModbusClient instance and connect
# External board for ttl to serial, connected to solar charge controller
modbus = ModbusClient(method='rtu', port='/dev/ttyAMA0', baudrate=9600,stopbits = 1, bytesize = 8, parity = 'N', timeout = 5, unit = 1) 
a = modbus.connect()
#print a

# Initialise database in public/downloadable location, create table if doesn't exist
dbCreate = sqlite3.connect('/var/www/html/SolarLog.db')
cursorCreate = dbCreate.cursor()
cursorCreate.execute('''CREATE TABLE IF NOT EXISTS Log ('0' INTEGER, '1' INTEGER, '2' INTEGER, '3' INTEGER, '4' INTEGER, '5' INTEGER, '6' INTEGER, '7' INTEGER, '8' INTEGER, '9' INTEGER, '10' INTEGER, '11' INTEGER, '12' INTEGER, '13' INTEGER, '14' INTEGER, '15' INTEGER, '16' INTEGER, '17' INTEGER, '18' INTEGER, '19' INTEGER, '20' INTEGER, '21' INTEGER, '22' INTEGER, '23' INTEGER, '24' INTEGER, '25' INTEGER, '26' INTEGER, '27' INTEGER, '28' INTEGER, '29' INTEGER, '30' INTEGER, '31' INTEGER, '32' INTEGER, '33' INTEGER, '34' INTEGER, LOG_DATE DATE, '35' INTEGER, '36' INTEGER, '37' INTEGER, '38' DATE, '39' INTEGER, '40' INTEGER)''')
dbCreate.commit()
dbCreate.close()
db = sqlite3.connect('/var/www/html/SolarLog.db')
cursor = db.cursor()
count = 0
backupCount = 0
directCount = 0
while 1:
	try:
		# Start constructing html for webpage
		html = '<html lang="en-US"><head><meta http-equiv="refresh" content="10"><meta name="viewport" content="width=device-width, initial-scale=1"></head>'
		html += '<style type="text/css">body { font-family: Verdana; font-size: 20pt; } table { font-family: Verdana; font-size: 20pt; } td { font-family: Verdana; font-size: 10pt;}</style><body>'

		# Read registers from the controller
		r = modbus.read_holding_registers(256, 35, unit=1)

		now = datetime.datetime.now()
		# Reset values we use to track Ah consumed through the shunt at the same time the solar controller resets its disharge Ah register
		if float(r.registers[18]) == 0 and prevDischargeAh > 0:
			directTotalAh = 0.0
			directTotalWh = 0.0
			directMaxAmps = 0.0
			elapsed = 0.0
			# It is now night time as far as rge solar controller is concerned
			daytime = False
		else:
			done = time.time()
			elapsed = done - start
		# Solar controller believes it is daytime. Now that daytime is here we can start updating charge status LEDs
		if float(r.registers[17]) == 0 and prevChargeAh > 0 or (prevChargeAh > 0 and float(r.registers[17]) > prevChargeAh):
			daytime = True

		# Used to track whether it's just become day or night according to the solar controller
		prevDischargeAh = float(r.registers[18])
		prevChargeAh = float(r.registers[17])

		# Calculate the shunts direct current and Ah used
		currentAh = (directAmps*(float(elapsed)/3600))
		directTotalAh += currentAh
		start = time.time()
		directTotalWh += (currentAh * (float(r.registers[1])/10))

		# Read amps from the shunt and set our variables
        adcDiff = adc.read_adc_difference(3, gain=GAIN)
		# adcDiff = 0
		directAmps = ((float(abs(adcDiff))*256.0)/32768)*1.333
		# directAmps = 1.2

		if directAmps > directMaxAmps:
			directMaxAmps = directAmps

		output = '<table border="0">'
		output += '<tr><td colspan="2">-------------------------------------------------------------------------------------------</td></tr>'
		output += '<tr><td colspan="2">------------------------------ SOLAR LIVE DATA ------------------------------</td></tr>'
		output += '<tr><td colspan="2">-------------------------------------------------------------------------------------------</td></tr>'
	    output += '<tr><td width="130">Battery:</td><td>' + str(float(r.registers[1])/10) + ' Volts, ' + str(r.registers[0]) + '% SOC</td></tr>'
		output += '<tr><td>Charge:</td><td>' + str(float(r.registers[2])/100) + ' Amps, '
		# Current charge status seems to be based on this register and these values
		if r.registers[32] == 32768:
			output += 'NORMAL'
		if r.registers[32] == 32770:
			output += 'MPPT'
		if r.registers[32] == 32771:
			output += 'EQUALISE'
		if r.registers[32] == 32772:
			output += 'BOOST'
		if r.registers[32] == 32773:
			output += 'FLOAT'
		output += '</td></tr>'
		output += '<tr><td>Solar:</td><td>' + str(float(r.registers[7])/10) + ' Volts, ' + str(float(r.registers[8])/100) + ' Amps</td></tr>'
	    output += '<tr><td>Load (cntrl/dir):</td><td>' + str(round((directAmps + float(r.registers[5])/100),1)) + ' (' + str(round(float(r.registers[5])/100,2)) + '/' + str(round(directAmps,2)) + ') Amps, ' + str(int(float(r.registers[6]) + (float(r.registers[1])/10 * float(directAmps)))) + ' Watts</td></tr>'
		# Temperatures seem to be based on a single hex value in the register, split in two and then converted to decimal
		output += '<tr><td>Temperature:</td><td>' + str(int(hex(r.registers[3])[2:-2],16)) + '&deg;C Device, ' + str(int(hex(r.registers[3])[-2:],16)) + '&deg;C Battery</td></tr>'
		output += '<tr><td></td></tr>'
		output += '<tr><td colspan="2">-------------------------------------------------------------------------------------------</td></tr>'
		output += '<tr><td colspan="2">--------------------------------- DAILY DATA -------------------------------------</td></tr>'
		output += '<tr><td colspan="2">-------------------------------------------------------------------------------------------</td></tr>'
		output += '<tr><td>Battery Min - Max:</td><td>' + str(float(r.registers[11])/10) + ' - ' + str(float(r.registers[12])/10) + ' Volts</td></tr>'
	    output += '<tr><td>Maximum Charge:</td><td>' + str(float(r.registers[13])/100) + ' Amps, ' + str(r.registers[15]) + ' Watts</td></tr>'
	    output += '<tr><td>Maximum Discharge:</td><td>' + str(float(r.registers[14])/100) + ' Amps, ' + str(r.registers[16]) + ' Watts</td></tr>'
		output += '<tr><td>Maximum (direct):</td><td>' + str(round(directMaxAmps,2)) + ' Amps, ' + str(round(float(r.registers[1])/10 * float(directMaxAmps),2)) + ' Watts</td></tr>'
	    output += '<tr><td>Charge Amount:</td><td>' + str(r.registers[17]) + ' Ah, ' + str(float(r.registers[19])/1000) + ' KWh</td></tr>'
	    output += '<tr><td>Discharge (cntrl/dir):</td><td>' + str(round(r.registers[18] + directTotalAh, 2)) + ' (' + str(r.registers[18]) + '/' + str(round(directTotalAh, 2))  + ') Ah, ' + str(round((float(r.registers[20])/1000) + (directTotalWh/1000), 2)) + ' KWh</td></tr>'
		# output += '<tr><td>Discharge (direct):</td><td>' + str(round(directTotalAh,3)) + ' Ah, ' + str(round((directTotalWh/1000),3)) + ' KWh</td></tr>'
		output += '<tr><td></td></tr>'
		output += '<tr><td colspan="2">-------------------------------------------------------------------------------------------</td></tr>'
		output += '<tr><td colspan="2">--------------------------------- GLOBAL DATA ---------------------------------</td></tr>'
		output += '<tr><td colspan="2">-------------------------------------------------------------------------------------------</td></tr>'
		output += '<tr><td>Fully Charged:</td><td>' + str(r.registers[23]) + ' of ' + str(r.registers[21]) + ' Days</td></tr>'
	    output += '<tr><td>Charge Amount:</td><td>' + str(r.registers[25]) + ' Ah, ' + str(float(r.registers[29])/1000) + ' KWh</td></tr>'
	    output += '<tr><td>Discharge Amount:</td><td>' + str(r.registers[27]) + ' Ah, ' + str(float(r.registers[31])/1000) + ' KWh</td></tr>'
	    	
		output += '<tr><td colspan="2">-------------------------------------------------------------------------------------------</td></tr>'
		# Only show the last router reboot if we have a value for it
		if lastInternetReset > datetime.datetime(1900, 1, 1, 1, 1, 1):
			output += '<tr><td>Last Router Reset:</td><td>' + str(lastInternetReset.strftime("%Y/%m/%d %H:%M:%S")) + '</td></tr>'
		output += '<tr><td>No Internet Count:</td><td>' + str(noInternetCount) + ' / ' + str(maxNoInternet) + '</td></tr>' 
		output += '<tr><td>Updated:</td><td>' + str(datetime.datetime.now().strftime("%Y/%m/%d %H:%M:%S")) + '</td></tr>'
        output += '<tr><td valign="top"><a href="SolarLogBackup.db">Download database</a></td>'
		# Unsure what these registers are for, just list at the bottom
	    output += '<td>' + str(r.registers) + '<!--</td></tr>' 
		output += '<tr><td colspan="2">--> ' + str(r.registers[4]) + ','
	    output += ' ' + str(r.registers[9]) + ','
	    output += ' ' + str(r.registers[10]) + ','
	    output += ' ' + str(r.registers[26]) + ','
	    output += ' ' + str(r.registers[24]) + ','
	    output += ' ' + str(r.registers[22]) + ','
	    output += ' ' + str(r.registers[28]) + ','
	    output += ' ' + str(r.registers[30]) + ','
	    output += ' ' + str(r.registers[33]) + ','
	    output += ' ' + str(r.registers[34]) + ','
		if daytime:
			output += ' DAY'
		else:
			output += ' NIGHT'
		output += '</td></tr></font>'
		html += output
		html += '</body></html>'
		# Write to solar.html where Apache web root is located
		f = open('/var/www/html/solar.html', 'w')
		f.write(html)
        	f.close()

		if count >= 3:
			# Only set charge status LEDs (on 5,6,13,19,26) if it's daytime
			if daytime:
				# Set the charge status LEDs
				if float(r.registers[17]) > (float(r.registers[18]) + directTotalAh + blueLED):
					# Blue
					GPIO.output(26, False)
					GPIO.output(19, False)
					GPIO.output(13, False)
					GPIO.output(6, False)
					GPIO.output(5, True)
				elif float(r.registers[17]) > (float(r.registers[18]) + directTotalAh + greenLED):
					# Green
					GPIO.output(26, False)
					GPIO.output(19, False)
					GPIO.output(13, False)
					GPIO.output(6, True)
					GPIO.output(5, False)
				elif float(r.registers[17]) > (float(r.registers[18]) + directTotalAh + whiteLED):
					# White
					GPIO.output(26, False)
					GPIO.output(19, False)
					GPIO.output(13, True)
					GPIO.output(6, False)
					GPIO.output(5, False)
				elif float(r.registers[17]) > (float(r.registers[18]) + directTotalAh + yellowLED):
					# Yellow
					GPIO.output(26, False)
					GPIO.output(19, True)
					GPIO.output(13, False)
					GPIO.output(6, False)
					GPIO.output(5, False)
				else:
					# Red
					GPIO.output(26, True)
					GPIO.output(19, False)
					GPIO.output(13, False)
					GPIO.output(6, False)
					GPIO.output(5, False)
			# Ping google for connectivity, if we haven't had internet for a while then switch relay connected to pin 23
			response = os.system("ping -c 1 -W 1 " + hostname + " > /dev/null")
			if response == 0:
				noInternetCount = 0
			else:
				noInternetCount += 1
			if noInternetCount == maxNoInternet:
				lastInternetReset = datetime.datetime.now()
				GPIO.output(23,False)
				time.sleep(15)
				GPIO.output(23,True)
			date = (time.strftime('%Y-%m-%d ') + time.strftime("%H:%M:%S"))
			# Save registers and ADS1115 readings to db
			cursor.execute('''INSERT INTO Log VALUES (?,?,?,?,?,?,?,?,?,?,?,?,?,?,?,?,?,?,?,?,?,?,?,?,?,?,?,?,?,?,?,?,?,?,?,?,?,?,?,?,?,?)''', (int(r.registers[0]), int(r.registers[1]), int(r.registers[2]), int(r.registers[3]), int(r.registers[4]), int(r.registers[5]), int(r.registers[6]), int(r.registers[7]), int(r.registers[8]), int(r.registers[9]), int(r.registers[10]), int(r.registers[11]), int(r.registers[12]), int(r.registers[13]), int(r.registers[14]), int(r.registers[15]), int(r.registers[16]), int(r.registers[17]), int(r.registers[18]), int(r.registers[19]), int(r.registers[20]), int(r.registers[21]), int(r.registers[22]), int(r.registers[23]), int(r.registers[24]), int(r.registers[25]), int(r.registers[26]), int(r.registers[27]), int(r.registers[28]), int(r.registers[29]), int(r.registers[30]), int(r.registers[31]), int(r.registers[32]), int(r.registers[33]), int(r.registers[34]), date, int(directAmps*100), int(directTotalAh*1000), int(noInternetCount), lastInternetReset, int(directMaxAmps*100),0))
			db.commit()
			# Reset no internet count if we just rebooted it, make negative to allow additional time to connect after reboot
			if noInternetCount == maxNoInternet:
				noInternetCount = -10
			count = 0
		else:
			count += 1

		#Copy to backup database file roughly every hour
		if backupCount == 600:
			copyfile('/var/www/html/SolarLog.db','/var/www/html/SolarLogBackup.db')
			backupCount = 0
		else:
			backupCount += 1

		time.sleep(5)
		# time.sleep(1)

	except Exception as e:
		html = str(e)
		html += 'Error, retrying in 2 seconds'
		html += '</body></html>'
		print html
		time.sleep(2)
	finally:
		modbus.close()
db.close()
